#!/usr/bin/env python
# tm5_tools.py

"""
Author : peters 

Revision History:
File created on 09 Feb 2009.
Major modifications to go to a class-based approach, July 2010.

This module holds specific functions needed to use the TM5 model within the data assimilation shell. It uses the information 
from the DA system in combination with the generic tm5.rc files. 

The TM5 model is now controlled by a python subprocess. This subprocess consists of an MPI wrapper (written in C) that spawns
a large number ( N= nmembers) of TM5 model instances under mpirun, and waits for them all to finish.

The design of the system assumes that the tm5.x (executable) was pre-compiled with the normal TM5 tools, and is residing in a 
directory specified by the ${RUNDIR} of a tm5 rc-file. This tm5 rc-file name is taken from the data assimilation rc-file. Thus,
this python shell does *not* compile the TM5 model for you!

"""

import os, sys
import subprocess
import logging
import shutil

sys.path.append(os.getcwd())

identifier      = 'TM5'
version         = 'release 3.0'
mpi_shell_filename  = 'tm5_mpi_wrapper'
mpi_shell_location  = 'da/bin/'

needed_rc_items = [
                    'rundir',
                    'outputdir',
                    'time.start',
                    'time.break.nday'
                  ]

################### Begin Class TM5 ###################

from da.baseclasses.observationoperator import ObservationOperator

class TM5ObservationOperator(ObservationOperator):
    """ This class holds methods and variables that are needed to run the TM5 model. It is initiated with as only argument a TM5 rc-file
        location. This rc-file will be used to figure out the settings for the run. 
        
        *** This method of running TM5 assumes that a pre-compiled tm5.exe is present, and it will be run from time.start to time.final ***
        
        These settings can be modified later. To run a model version, simply compile the model using an existing TM5 rc-file, then 
        open python, and type:

           []> tm=TM5('/Users/peters/Modeling/TM5/tutorial.rc')
           []> tm.WriteRc()
           []> tm.WriteRunRc()
           []> tm.Run()

        To use this class inside a data assimilation cycle, a stand-alone method "Initialize()" is included which modifies the TM5
        settings according to an external dictionary of values to overwrite, and then runs the TM5 model.

    
    """

    def __init__(self):
        """ The instance of an TMObservationOperator is application dependent """
        self.Identifier     = self.getid()    # the identifier gives the model name
        self.Version        = self.getversion()       # the model version used

        msg                 = 'Observation Operator object initialized: %s (%s)'%(self.Identifier, self.Version,) ; logging.info(msg)

    def getid(self):
        return identifier

    def getversion(self):
        return version

    def Initialize(self, DaCycle):
        """ 
        Prepare a forward model TM5 run, this consists of:

          - reading the TM5 rc-file, 
          - validating it, 
          - modifying the values,
          - Creating a tm5_runtime.rc file
          - Removing the existing tm5.ok file if present
    
        """
        from da.tools.general import CreateLinks

        RcFileName          = DaCycle.da_settings['forecast.model.rc']
        dummy               = self.LoadRc(RcFileName)   # load the specified rc-file
        dummy               = self.ValidateRc()         # validate the contents

# Create a link from TM5 to the rundirectory of the das system

        sourcedir           = self.tm_settings['rundir']
        targetdir           = os.path.join(DaCycle.da_settings['dir.exec'],'tm5')
        dummy               = CreateLinks(sourcedir,targetdir)

        DaCycle.da_settings['dir.exec.tm5'] = targetdir

# Write a modified TM5 model rc-file in which run/break times are defined by our da system

        NewItems = {
                    'time.start'    : DaCycle.da_settings['time.sample.start']    ,
                    'time.final'    : DaCycle.da_settings['time.sample.end']    ,
                    'rundir'        : DaCycle.da_settings['dir.exec.tm5'] ,
                    'outputdir'     : DaCycle.da_settings['dir.output'] ,
                    'savedir'       : DaCycle.da_settings['dir.save']    ,
                    'das.input.dir' : DaCycle.da_settings['dir.input']
                    }

        if DaCycle.da_settings['time.restart']:  # If this is a restart from a previous cycle, the TM5 model should do a restart
            NewItems['istart'] = 3
        if DaCycle.da_settings['time.sample.window'] != 0:  # If this is a restart from a previous time step wihtin the filter lag, the TM5 model should do a restart
            NewItems['istart'] = 3
        
        # If neither one is true, simply take the istart value from the tm5.rc file that was read

        self.ModifyRC(NewItems)

        self.WriteRc()

        self.WriteRunRc()

# Copy the pre-compiled MPI wrapper to the execution directory

        targetdir = os.path.join(DaCycle.da_settings['dir.exec.tm5'])

        if not os.path.exists(os.path.join(mpi_shell_location,mpi_shell_filename) ):
            msg = "Cannot find the mpi_shell wrapper needed for completion (%s) in (%s)"% (mpi_shell_filename,mpi_shell_location) ; logging.error(msg)
            msg = "Please see the %s/readme_wrapper.txt file for instructions to compile it"%mpi_shell_location ; logging.error(msg)
            raise IOError

        shutil.copy(os.path.join(mpi_shell_location,mpi_shell_filename) ,os.path.join(targetdir,mpi_shell_filename) )

        return 0

    def LoadRc(self,RcFileName):
        """ 
        This method loads a TM5 rc-file with settings for this simulation 
        """
        import da.tools.rc as rc

        self.tm_settings    = rc.read(RcFileName)
        self.RcFileName     = RcFileName
        self.Tm5RcLoaded    = True

        msg                 = 'TM5 rc-file loaded successfully' ; logging.debug(msg)

        return True

    def ValidateRc(self):
        """
        Validate the contents of the tm_settings dictionary and add extra values. The required items for the TM5 rc-file
        are specified in the tm5_tools module, as dictionary variable "needed_rc_items".
        """
        from da.tools.general import ToDatetime

        for k,v in self.tm_settings.iteritems():
            if v == 'True' : self.tm_settings[k] = True
            if v == 'False': self.tm_settings[k] = False
            if 'date' in k : self.tm_settings[k] = ToDatetime(v)
            if 'time.start' in k : 
                self.tm_settings[k]              = ToDatetime(v,fmt='TM5')
            if 'time.final' in k : 
                self.tm_settings[k]              = ToDatetime(v,fmt='TM5')

        for key in needed_rc_items:

            if not self.tm_settings.has_key(key):
                status,msg = ( False,'Missing a required value in rc-file : %s' % key)
                logging.error(msg)
                raise IOError,msg

        status,msg = ( True,'rc-file has been validated succesfully' )  ; logging.debug(msg)




    def ModifyRC(self,NewValues):
        """ 
        Modify parts of the tm5 settings, for instance to give control of file locations to the DA shell
        instead of to the tm5.rc script. 

        Note that we replace these values in all {key,value} pairs of the tm5.rc file!

        """
    
        for k,v in NewValues.iteritems():
            if self.tm_settings.has_key(k):
                # keep previous value
                v_orig= self.tm_settings[k]
                #replace with new
                self.tm_settings[k] = v
                #replace all instances of old with new, but only if it concerns a name of a path!!!
                if os.path.exists(str(v)): 
                    for k_old,v_old in self.tm_settings.iteritems():
                        if not isinstance(v_old,str): 
                            continue
                        if str(v_orig) in str(v_old): 
                            v_new = str(v_old).replace(str(v_orig),str(v))
                            self.tm_settings[k_old] = v_new

                logging.debug('Replaced tm5 rc-item %s ' % k)

            else:
                self.tm_settings[k] = v
                logging.debug('Added new tm5 rc-item %s ' % k)


    def WriteRc(self):
        """ 
        Write the rc-file settings to a tm5.rc file in the rundir
        """
        import da.tools.rc as rc

        tm5rcfilename   = os.path.join(self.tm_settings['rundir'],'tm5.rc')

        dummy           = rc.write(tm5rcfilename,self.tm_settings)

        msg             = "Modified tm5.rc written (%s)" % tm5rcfilename      ;logging.debug(msg)

    def WriteRunRc(self):
        """ 
        Create the tm5-runtime.rc file which is read by initexit.F90 to control time loop and restart from save files
        """
        import da.tools.rc as rc

        tm5rcfilename   = os.path.join(self.tm_settings['rundir'],'tm5_runtime.rc')

        dummy           = rc.write(tm5rcfilename,self.tm_settings)

        rc_runtm5={}

        rc_runtm5['year1']  = self.tm_settings['time.start'].year
        rc_runtm5['month1'] = self.tm_settings['time.start'].month
        rc_runtm5['day1']   = self.tm_settings['time.start'].day
        rc_runtm5['hour1']  = self.tm_settings['time.start'].hour
        rc_runtm5['minu1']  = 0
        rc_runtm5['sec1']   = 0

        rc_runtm5['year2']  = self.tm_settings['time.final'].year
        rc_runtm5['month2'] = self.tm_settings['time.final'].month
        rc_runtm5['day2']   = self.tm_settings['time.final'].day
        rc_runtm5['hour2']  = self.tm_settings['time.final'].hour
        rc_runtm5['minu2']  = 0
        rc_runtm5['sec2']   = 0

        rc_runtm5['istart']    = self.tm_settings['istart']
        rc_runtm5['savedir']   = self.tm_settings['savedir']
        rc_runtm5['outputdir'] = self.tm_settings['outputdir']

        rc.write(tm5rcfilename,rc_runtm5)

        msg             = "Modified tm5_runtime.rc written (%s)" % tm5rcfilename      ;logging.debug(msg)


    def ValidateInput(self,DaCycle):
        """
        Make sure that parameter files are written to the TM5 inputdir, and that observation lists are present
        """

        datadir   = self.tm_settings['das.input.dir']
        if not os.path.exists(datadir):
            msg   = "The specified input directory for the TM5 model to read from does not exist (%s), exiting..."%datadir ; logging.error(msg)
            raise IOError,msg

        datafiles = os.listdir(datadir)

        obsfile   = DaCycle.da_settings['ObsOperator.inputfile']

        if not os.path.exists(obsfile):
            msg   = "The specified obs input file for the TM5 model to read from does not exist (%s), exiting..."%obsfile ; logging.error(msg)
            raise IOError,msg


        for n in range(int(DaCycle.da_settings['forecast.nmembers'])):
            paramfile = 'parameters.%03d.nc'%n
            if paramfile not in datafiles:
                msg = "The specified parameter input file for the TM5 model to read from does not exist (%s), exiting..."%paramfile ; logging.error(msg)
                raise IOError,msg

        # Next, make sure there is an actual model version compiled and ready to execute

        targetdir          = os.path.join(self.tm_settings['rundir'])
        self.Tm5Executable = os.path.join(targetdir,'tm5.x')
        if not os.path.exists(self.Tm5Executable):
            msg = "Required TM5 executable was not found %s"%self.Tm5Executable                           ; logging.error(msg)
            msg = "Please compile the model with the specified rc-file and the regular TM5 scripts first" ; logging.error(msg)
            raise IOError

        return 0

    def Run(self,DaCycle):
        """ 
         Start the TM5 executable. A new log file is started for the TM5 model IO, and then a subprocess is
         spawned with the tm5_mpi_wrapper and the tm5.x executable. The exit code of the model is caught and
         only if successfull on all processors will execution of the shell continue. 
         
        """

        cwd = os.getcwd()

        # From here on, several options should be implemented.

        #
        # (1) Where an mpi process is forked to do a TM5 instance with N tracers, each an ensemble member
        #
        # (2) Where N processes are spawned, each being one TM5 instance representing one member
        #
        # (3) Where N/m processes are spawned, each being a TM5 instance that handles m ensemble members
        #
        # In principle, it is best to make these processes produce scripts that can be executed stand-alone, or
        # be submitted to a queue.
        #

        # Open logfile and spawn model, wait for finish and return code

        # Code for Option (1)

        code = self.TM5_under_mpirun(DaCycle)

        if code == 0:
            logging.info('Finished model executable succesfully (%s)'%code)
            self.Status = 'Success'
        else:
            logging.error('Error in model executable return code: %s ' % code)
            self.Status = 'Failed'
            raise OSError

        # Return to working directory

        os.chdir(cwd)

        return code

    def TM5_under_mpirun(self, DaCycle):
        """ Method handles the case where a shell runs an MPI process that forks into N TM5 model instances """
        from string import join
        import datetime

        DaPlatForm              = DaCycle.DaPlatForm

        targetdir               = os.path.join(self.tm_settings['rundir'])

        # Go to executable directory and start the subprocess, using a new logfile

        dummy                   = os.chdir(targetdir)
        msg                     = 'Changing directory to %s ' % targetdir       ;logging.debug(msg)

        # Remove the tm5.ok file from a previous run, placed back only if a successful TM5 run is executed

        okfile                  = 'tm5.ok'
        if os.path.exists(okfile): 
            dummy               = os.remove(okfile)

        nprocesses              = DaCycle.da_settings['forecast.nmembers']
        jobparams               = {'jobname':'tm5',
                                   'jobnodes':'ncomp %d'%int(nprocesses),
                                   'jobtime':'00:30:00',
                                   'joblog':os.path.join(DaCycle.da_settings['dir.jobs'])
                                  }

        template                = DaPlatForm.GetJobTemplate(jobparams,block=True)
        template                += 'cd %s\n'%targetdir
        template                += 'mpirun -np %d %s ./tm5.x\n'%(int(nprocesses),mpi_shell_filename,)

        jobfile                 = DaPlatForm.WriteJob(DaCycle,template,'tm5')
      
        msg                     = 'Submitting job at %s'%datetime.datetime.now() ; logging.debug(msg)
        code                    = DaPlatForm.SubmitJob(jobfile) 
        msg                     = 'Resuming job at %s'%datetime.datetime.now() ; logging.debug(msg)
        
        if not os.path.exists(okfile): 
            code = -1
        else:
            code = 0

        return code


    def TM5_With_N_tracers(self, nprocesses):
        """ Method handles the case where one TM5 model instance with N tracers does the sampling of all ensemble members"""
        from string import join

        modellogfile             =   open(self.ModelLogFilename,'w')
        #cmd    = ['openmpirun','-np', '10', mpi_shell_filename,'./tm5.x']
        cmd     = ['mpirun','-np',str(nprocesses),'./tm5.x']
        msg     = 'Starting model executable as subprocess (%s)'%join(cmd,' ') ; logging.info(msg)

        #code    = subprocess.call(cmd,stdout=modellogfile,stderr=modellogfile)

        modellogfile.close()

    def SaveData(self):
        """ Write the TM5 recovery data for the next cycle """

        sourcedir   = os.path.join(self.tm_settings['outputdir'])
        targetdir   = os.path.join(self.tm_settings['savedir'])
        filter      = ['save_%s'%self.tm_settings['time.final'].strftime('%Y%m%d')]

        msg         = "Performing a full backup of TM5 save data"                                 ; logging.debug(msg)
        msg         = "           from directory: %s " % sourcedir                                ; logging.debug(msg)
        msg         = "           to   directory: %s " % targetdir                                ; logging.debug(msg)
        msg         = "           with filter: %s " % filter                                      ; logging.debug(msg)

        for file in os.listdir(sourcedir):

            file = os.path.join(sourcedir,file)

            if os.path.isdir(file): # skip dirs
                skip = True
            elif filter == []:      # copy all
                skip= False        
            else:                   # check filter
                skip = True         # default skip
                for f in filter:
                    if f in file: 
                        skip = False # unless in filter
                        break
                
            if skip: 
                msg         = "           [skip] .... %s " % file                                     ; logging.debug(msg)
                continue    

            msg         = "           [copy] .... %s " % file                                                ; logging.debug(msg)
            dummy       = shutil.copy(file,file.replace(sourcedir,targetdir) )

################### End Class TM5 ###################
                   

if __name__ == "__main__":

    from da.tools.initexit import StartLogger
    from da.tools.pipeline import JobStart
    import datetime as dtm
    import sys,os

    sys.path.append(os.getcwd())

    StartLogger()

    DaCycle = JobStart([],{'rc':'da.rc'})
    DaCycle.Initialize()
    DaCycle.da_settings['time.sample.start'] = dtm.datetime(2000,1,1)
    DaCycle.da_settings['time.sample.end'] = dtm.datetime(2000,1,2)
    DaCycle.da_settings['time.sample.window'] = 0

    tm=TM5ObservationOperator()
    tm.Initialize(DaCycle)
    tm.Run(DaCycle)
    tm.SaveData()





